
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "ob_inter.h"

#define TFACTR 0.80
#define STOP   0.20
#define CB     30.00

#define RANRES (0xFFF)
#define TTRIALS 20

typedef void (*ob_neigh_t) (Interval *);

ob_neigh_t ob_generate_neigh;

static double tfactr;
static double stop;
static double cb;
static int ttrials;

static int metrop( double, double);
static double calcRelative( double, double );

static void fullRandomSelection		      (Interval *);		
static void partialRandomSelection	      (Interval *);		
static void randomDistanceSelection	      (Interval *);		
static void fullRatedSelection		      (Interval *);		
static void partialRatedSelection	      (Interval *);		
static void temperatureDistanceSelection      (Interval *);	
static void smallRandomDistanceSelection      (Interval *);	
static void smallTemperatureDistanceSelection (Interval *);	
static void treeRandomSelection		      (Interval *);		
static void temperatureTreeRandomSelection    (Interval *);    


static inline int left(int);
static inline int right(int);
static inline int parent(int);



void printStatistics(void);

/*--------------------------------------------------------*/
/* For statistics */
/*--------------------------------------------------------*/

static int statistics_total_invocations = 0;
static int statistics_temperature_changes = 0;
static int statistics_improvements_recorded = 0;

static inline double
dran(void){
   double n = (double)((random() & RANRES)/(double)RANRES );
   return n <= 0.001 ? 0.01 : n ;
}

/*******************************************************************/

void
printStatistics(void){
   printf("Total invocations:     %d\n",statistics_total_invocations);
   printf("Temperature changes:   %d\n",statistics_temperature_changes);
   printf("Improvements recorded: %d\n",statistics_improvements_recorded);
}


/*******************************************************************/

static int
metrop( double relativeCost, double temp ){
   double ran = ((double)((random() & RANRES)/RANRES ));
		 
   return relativeCost < 0 || ran < ( exp( -(relativeCost*cb)/temp));
}

/*******************************************************************/

static inline double
calcRelative( double t1, double t2 ){
   return (MAX_TIME == t1 ) ? -0.001 : ( t2 - t1 ) / t1 ;
}


/*******************************************************************/

static void
swap( void* array , int i1, int i2, int tsize ){
   char t[12], *a;
   a = (char*)array;
   i1 *= tsize;
   i2 *= tsize;
   memcpy( t, a+i1 , tsize );
   memcpy( a+i1, a+i2, tsize );
   memcpy( a+i2, t, tsize );
}

/*******************************************************************/
/* NEIGHBOURHOOD FUNCTIONS */
/*******************************************************************/

static void
fullRandomSelection( Interval *iv ){
   int i;
   Info *ifo;
   for( i = 0; i < iv -> ilength ; i ++ ){
      ifo = iv -> info[i];
      ifo -> cidx = random() % ifo -> selectionLength;
   }
}

static void
partialRandomSelection( Interval *iv ){
   int i;
   Info *ifo;
   for( i = 0; i < iv -> ilength ; i ++ ){
      if( (random() & 0xf) < 0xa ){
	 ifo = iv -> info[i];
	 ifo -> cidx = random() % ifo -> selectionLength;
      }
   }
}

static void
randomDistanceSelection( Interval *iv ){
   int i,maxmove,dlen;
   Info *ifo;
   
   for( i = 0; i < iv -> ilength ; i ++ ){
      ifo = iv -> info[i];
      dlen = ifo -> selectionLength / 3;
      dlen = dlen ? dlen : 1 ;
      maxmove = random() % dlen ;
      ifo -> cidx = ifo -> aidx;
      ifo -> cidx += random() & 1 ? maxmove : -maxmove ;
      if( ifo -> cidx < 0 || ifo -> cidx >= ifo -> selectionLength )
	 ifo -> cidx = ifo -> aidx;
   }
}

static void
fullRatedSelection( Interval *iv ){
   int i;
   Info *ifo;
   
   for( i = 0; i < iv -> ilength ; i ++ ){
      ifo = iv -> info[i];
      ifo -> cidx = (int)(exp(log(dran())/iv->temp) \
			  * ifo -> selectionLength);
   }
}

static void
partialRatedSelection( Interval *iv ){
   int i;
   Info *ifo;
   
   for( i = 0; i < iv -> ilength ; i ++ ){
      if( (random() & 0xf) > 0xa ) continue;
      ifo = iv -> info[i];
      ifo -> cidx = (int)(exp(log(dran())/iv->temp) \
			  * ifo -> selectionLength);
   }
}

static void
temperatureDistanceSelection( Interval *iv ){
   int i,m,l;
   Info *ifo;
   
   for( i = 0; i < iv -> ilength ; i ++ ){
      ifo = iv -> info[i];
      m = -1;
      if( exp(log(dran())/iv->temp) > 0.5 ){
	 m = 1;
      }
      l = (int)(exp(log(dran())/iv->temp) \
		* ifo -> selectionLength);

      ifo -> cidx = ifo -> aidx + (m * l);
      if( ifo -> cidx < 0 || ifo -> cidx >= ifo -> selectionLength )
	 ifo -> cidx = ifo -> aidx ;
   }
}

static void
smallRandomDistanceSelection( Interval *iv ){
   int i;
   Info *ifo;

   for( i = 0; i < iv -> ilength ; i ++ ){
      ifo = iv -> info[i];
      ifo -> cidx = ifo -> aidx;
      ifo -> cidx += random() & 1 ? 1 : -1 ;
      if( ifo -> cidx < 0 || ifo -> cidx >= ifo -> selectionLength )
	 ifo -> cidx = ifo -> aidx;
   }   
}

static void
smallTemperatureDistanceSelection( Interval *iv ){
   int i,m;
   Info *ifo;
   
   for( i = 0; i < iv -> ilength ; i ++ ){
      ifo = iv -> info[i];
      m = -1;
      if( exp(log(dran())/iv->temp) > 0.5 ){
	 m = 1;
      }

      ifo -> cidx = ifo -> aidx + m;
      if( ifo -> cidx < 0 || ifo -> cidx >= ifo -> selectionLength )
	 ifo -> cidx = ifo -> aidx ;
   }   
}

static void
treeRandomSelection( Interval *iv ){
   int i;
   Info *ifo;

   for( i = 0; i < iv -> ilength ; i ++ ){
      ifo = iv -> info[i];
      ifo -> cidx = ifo -> aidx;
      ifo -> cidx = random() & 1 ? parent( ifo -> aidx ) : \
	 random () & 1 ? left( ifo -> aidx ) : right( ifo -> aidx ) ;

      if( ifo -> cidx < 0 || ifo -> cidx >= ifo -> selectionLength )
	 ifo -> cidx = ifo -> aidx;
   }      
}

static void
temperatureTreeRandomSelection( Interval *iv ){
   int i;
   Info *ifo;

   for( i = 0; i < iv -> ilength ; i ++ ){
      ifo = iv -> info[i];
      if(exp(log(dran())/iv->temp) < 0.5)
	 ifo -> cidx = parent(ifo -> aidx);
      else if( random() & 1 )
	 ifo -> cidx = left(ifo -> aidx);
      else
	 ifo -> cidx = right(ifo -> aidx );


      if( ifo -> cidx < 0 || ifo -> cidx >= ifo -> selectionLength )
	 ifo -> cidx = ifo -> aidx;
   }         
}

/*******************************************************************/

static inline int
left(int i){
   return (i<<1) + 1;
}

static inline int
right(int i){
   return (i<<1) +2;
}

static inline int
parent(int i){
   return (i-1)>>1;
}

static void
manageHeap( Interval *iv, double relativeCost ){
   int i,idx,max,vaidx,ceqa;
   Info *ifo;
   double *relcost;
   int *selection;
   double newCost;
   

   for( i = 0; i < iv -> ilength ; i ++ ){
      ifo = iv -> info[i];
      // set in new cost and update heap
      idx = ifo -> cidx;
      max = ifo -> selectionLength;
      if( !(idx == 0 && relativeCost > 0) )
	 newCost = ifo -> relcost[0] + relativeCost ;
      else
	 return;
      
      // only accept improvements
      // makes the downward traversal unnecessary
      if( newCost > ifo -> relcost[idx] )
	 if( ifo -> relcost[idx] != 0.00 )
	    return;

      ifo -> relcost[idx] = newCost ;
      relcost = ifo -> relcost;
      selection = ifo -> selection;

      // take care of the accepted system state: aidx
      ceqa = (ifo->cidx == ifo->aidx);
      vaidx = selection[ifo->aidx];
      
      if( idx )
	 while( relcost[idx] < relcost[parent(idx)] ){
	    swap( relcost, idx, parent(idx) , sizeof(double));
	    swap( selection, idx, parent(idx), sizeof(int));
	    idx = parent(idx);
	    if( 0 > parent(idx))
		break;
	 }
      if( idx < max ){
	 // swap with smallest
	 while(1){
	    int l,r,smallest = idx;
	    l = left(idx);
	    r = right(idx);
	
	    if( l < max && relcost[l] < relcost[idx] )
	       smallest = l;
	    if( r < max && relcost[r] < relcost[idx] )
	       smallest = r;
	    if( smallest != idx ){
	       swap( relcost, idx, smallest, sizeof(double));
	       swap( selection, idx, smallest, sizeof(int));
	       idx = smallest;
	    }
	    else
	       break;
	 }
      }

      ifo -> cidx = idx;
      if( ceqa )
	 ifo -> aidx = idx;
      else if( selection[ifo->aidx] != vaidx ){
	 int pval,lval,rval;
	 pval = parent(ifo->aidx);
	 lval = left(ifo->aidx);
	 rval = right(ifo->aidx);
	 
	 if( pval > -1 && pval < max )
	    pval = selection[pval];
	 if( lval > -1 && lval < max )
	    lval = selection[lval];
	 if( rval > -1 && rval < max )
	    rval = selection[rval];
	 
	 if( vaidx == pval )
	    ifo -> aidx = parent(ifo -> aidx);
	 if( vaidx == lval )
	    ifo -> aidx = left(ifo -> aidx);
	 if( vaidx == rval )
	    ifo -> aidx = right(ifo -> aidx);
      }
   }
}

/*******************************************************************/

void
ob_sao_init(void){
   char *env;
   ob_neigh_t neighbour_generators[10];

   neighbour_generators[0] = fullRandomSelection;
   neighbour_generators[1] = partialRandomSelection;
   neighbour_generators[2] = randomDistanceSelection;
   neighbour_generators[3] = fullRatedSelection;
   neighbour_generators[4] = partialRatedSelection;
   neighbour_generators[5] = temperatureDistanceSelection;
   neighbour_generators[6] = smallRandomDistanceSelection;
   neighbour_generators[7] = smallTemperatureDistanceSelection;
   neighbour_generators[8] = treeRandomSelection;
   neighbour_generators[9] = temperatureTreeRandomSelection;    

 
   if( (env = getenv("OB_NEIGHBOUR_GENERATOR")) != NULL)
      ob_generate_neigh = neighbour_generators[atoi(env)];
   else
      ob_generate_neigh = temperatureDistanceSelection;

   if((env = getenv("OB_SAO_TFACTR")) != NULL)
      tfactr = atof(env);
   else
      tfactr = TFACTR;
   if((env = getenv("OB_SAO_STOP")) != NULL)
      stop = atof(env);
   else
      stop = STOP;
   if((env = getenv("OB_SAO_CB")) != NULL)
      cb = atof(env);
   else
      cb = CB;
   if((env = getenv("OB_SAO_TTRIALS")) != NULL)
      ttrials = atoi(env);
   else
      ttrials = TTRIALS;
}


/*******************************************************************/


void 
ob_simAnnealOnce( Interval *iv, double itime, int idx ){
      double relativeCost;

      if( iv->temp < stop){
	 return;
      }

      if( iv->firstIteration ){
	 iv->firstIteration = 0;
	 return;
      }

      (iv -> trials) ++ ;
      statistics_total_invocations ++;
      
      relativeCost = calcRelative( iv->btimes[idx], itime );

      manageHeap( iv, relativeCost );
      
      if( metrop( relativeCost, iv -> temp)){
	 Info *ifo;     
	 int i;

	 iv->times[idx] = itime;
	 if( itime < (iv->btimes[idx]) ){
	    iv -> btimes[idx] = itime;
	    for( i = 0; i < (iv->ilength); i ++ ){
	       ifo = iv -> info[i];
	       ifo -> aidx = ifo -> bidx = ifo -> cidx;
	    }
	 }
	 else{
	    for( i = 0; i < (iv->ilength) ; i ++ ){
	       ifo = iv -> info[i];
	       ifo -> aidx = ifo -> cidx;
	    }
	 }
      }

      
      if( iv -> trials >= ttrials ){
	 iv -> trials = 0;
	 iv -> temp *= tfactr;
	 statistics_temperature_changes ++;
      }

      // create a new candidate
      ob_generate_neigh( iv );
}

